import numpy as np
import pandas as pd
import typing as tp

from .iid_bootstrap import iid_bootstrap
from .block_bootstrap import (
    moving_block_bootstrap,
    circular_block_bootstrap,
    stationary_bootstrap,
    BlockBootstrapType
)

from .tapered_block_bootstrap import (
    tapered_block_bootstrap,
    tapered_block_bootstrap_vectorized
)


def rolling_window(a, window):
    shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
    strides = a.strides + (a.strides[-1],)
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)


def calculate_number_of_complete_blocks(sample_length: int,
                                        block_length: int) -> int:
    """
    This function computes the number of blocks that fit into a sample of a
    given length.
    """

    return int(sample_length / block_length)


def _get_number_of_observations_from_1d_or_2d_array(a: np.ndarray) -> int:
    """
    This function retrieves the number of observations in a numpy array with
    shape (T, ) or (k, T), where k is the dimensionality of the data.

    Args:
        a: numpy data array

    Returns: number of observations

    """
    if a.ndim == 1:
        T = len(a)
    elif a.ndim == 2:
        _, T = a.shape
    else:
        raise ValueError('The data array must be a NumPy array with one or '
                         'two axes')

    return T


def remove_incomplete_blocks(log_returns: np.ndarray,
                             number_of_observations_to_aggregate: int) \
        -> np.ndarray:
    """
    This function takes a NumPy array of observations and a positive integer of
    observations to be aggregated. It cuts off the remaining observations at the
    end that do not form a complete block.
    Args:
        log_returns: a numpy with shape (T, ) or (k, T)
        number_of_observations_to_aggregate: the block length

    Returns: a numpy array with the remaining observations cut off

    """
    T \
        = _get_number_of_observations_from_1d_or_2d_array(log_returns)

    number_of_complete_blocks \
        = calculate_number_of_complete_blocks(
            T,
            number_of_observations_to_aggregate)
    end_index = number_of_complete_blocks * number_of_observations_to_aggregate

    if log_returns.ndim == 1:
        return log_returns[:end_index]
    elif log_returns.ndim == 2:
        return log_returns[:, :end_index]
    else:
        assert False


def aggregate_log_returns(log_returns: np.ndarray,
                          number_of_observations_to_aggregate: int,
                          rolling: bool = False,
                          verbose: bool = True) \
        -> tp.Tuple[np.ndarray, np.ndarray]:
    """
    Args:
        log_returns: a numpy array of log returns with shape (T, ) or (k, T),
                     where k is number of securities or portfolios
        number_of_observations_to_aggregate: the length of blocks to aggregate
        rolling: If True: the step size to move the summation window by is one
                          observation.
                 If False: the step size is number_of_observations_to_aggregate
                           resulting in sums of returns from non-overlapping
                           observations.
        verbose: Whether to print the number of observations that were cut off
                 from the sample at the end because they were not sufficient to
                 form another aggregation period.
                 This is 0 if rolling equals True.

    Returns:
        Tuple of (aggregated_log_returns, retained_log_returns).
        The aggregated_log_returns are the sums of
        'number_of_observations_to_aggregate' log returns.
    """

    if number_of_observations_to_aggregate < 1:
        raise ValueError("The number of observations to aggregate must be a "
                         "positive integer.")

    if number_of_observations_to_aggregate == 1:
        return log_returns, log_returns

    n_observations \
        = _get_number_of_observations_from_1d_or_2d_array(log_returns)

    if rolling:
        if log_returns.ndim == 1:
            log_returns_df = pd.Series(log_returns)
            aggregated_log_returns \
                = (log_returns_df
                   .rolling(number_of_observations_to_aggregate)
                   .sum()
                   .dropna()
                   .values)
        elif log_returns.ndim == 2:
            rolling_log_returns \
                = rolling_window(log_returns,
                                 number_of_observations_to_aggregate)
            aggregated_log_returns = rolling_log_returns.sum(axis=2)
        else:
            raise ValueError(
                "The argument log_returns must be a one- or "
                "two-dimensional array")
        retained_log_returns = log_returns
    else:
        number_of_complete_blocks \
            = calculate_number_of_complete_blocks(
                n_observations,
                number_of_observations_to_aggregate)
        end_index \
            = number_of_complete_blocks * number_of_observations_to_aggregate

        if log_returns.ndim == 1:
            # the input is a single time series sample
            retained_log_returns \
                = log_returns[:end_index]
            aggregated_log_returns \
                = (retained_log_returns
                   .reshape(number_of_complete_blocks,
                            number_of_observations_to_aggregate)
                   .sum(axis=1))
        elif log_returns.ndim == 2:
            # the input are many time series samples
            number_of_complete_blocks \
                = calculate_number_of_complete_blocks(
                    n_observations,
                    number_of_observations_to_aggregate)
            retained_log_returns \
                = log_returns[:, :end_index]

            aggregated_log_returns \
                = (retained_log_returns
                   .reshape(-1,
                            number_of_complete_blocks,
                            number_of_observations_to_aggregate)
                   .sum(axis=2))
        else:
            raise ValueError(
                'The argument log_returns must be a NumPy array with '
                'one or two axes')

    if verbose:
        number_of_cut_off_observations \
            = n_observations % number_of_observations_to_aggregate
        print(f'Number of cut off observations '
              f'= {number_of_cut_off_observations}')

    return aggregated_log_returns, retained_log_returns


def aggregate_multi_dimensional_log_returns(
        log_returns: np.ndarray,
        number_of_observations_to_aggregate: int):
    """
    This function takes a numpy array with shape (replications, T, dimensions)
    and sums the T observations in each replication and along each dimension in
    blocks. The function does not discard observations that do not fit a block.

    Args:
        log_returns: a numpy array with shape (replications, T, dimensions)
        number_of_observations_to_aggregate: number of observations to sum

    Returns: a numpy array with shape (replications,
                                       number_of_blocks,
                                       dimension)

    """
    replications, T, dimensions = log_returns.shape

    number_of_complete_blocks \
        = calculate_number_of_complete_blocks(
            sample_length=T,
            block_length=number_of_observations_to_aggregate)

    if number_of_complete_blocks * number_of_observations_to_aggregate != T:
        raise ValueError('T / number_of_observations_to_aggregate must be an '
                         'integer')

    aggregated_log_returns \
        = (log_returns
           .reshape((replications,
                    number_of_complete_blocks,
                    number_of_observations_to_aggregate,
                    dimensions))
           .sum(axis=2))

    return aggregated_log_returns


def _calculate_number_of_observations_to_be_cut_off_at_the_end(
        sample_length: int,
        number_of_observations_to_aggregate: int,
        verbose: bool = False) -> int:
    """
    This function supports the calculation of statistics from a single
    time-series by shifting the window by number_of_observations_to_aggregate
    different integer values to obtain different estimates for the same
    statistic.

    Specifically, the function calculates how many observations need to be
    discarded at the end if number_of_observations_to_aggregate different
    shifted windows are considered.

    Args:
        sample_length: the length of the time-series
        number_of_observations_to_aggregate: the block length
        verbose: whether to print debug information

    Returns:

    """
    if sample_length < 2 * number_of_observations_to_aggregate - 1:
        raise ValueError(
                f'The sample length is too short. At least '
                f'{2 * number_of_observations_to_aggregate - 1} observations'
                f' are required.')

    observations_remaining_at_the_end \
        = sample_length % number_of_observations_to_aggregate

    observations_to_be_cut_off_at_the_end \
        = (observations_remaining_at_the_end + 1) \
        % number_of_observations_to_aggregate

    if verbose:
        print(f'observations remaining at the end '
              f'= {observations_remaining_at_the_end}')
        print(observations_to_be_cut_off_at_the_end)

    return observations_to_be_cut_off_at_the_end


def resample_and_aggregate(
        log_returns: np.ndarray,
        block_length: tp.Union[int, float],
        replications: int,
        number_of_observations_to_aggregate: int,
        block_bootstrap_type: BlockBootstrapType
        = BlockBootstrapType.STATIONARY,
        sub_sample_length: int = None,
        rolling: bool = False,
        verbose: bool = True) -> tp.Tuple:
    """
    This function
    1.) computes aggregated returns from of an original time-series of
    log returns discarding insufficient observations
    2.) and draws new bootstrap samples from those original observations not
    discarded and aggregates these resampled returns

    Args:
        log_returns: a numpy array of log returns with one axis
        block_length: the block length for bootstrapping
        replications: the number of bootstrap samples to simulate
        number_of_observations_to_aggregate: the number of log returns to
                aggregate in a block
        block_bootstrap_type: The block bootstrap method to use
                (see class BlockBootstrapType)
        sub_sample_length: the length of each bootstrap sub-samples to generate
                (optional parameter which defaults to the same length as the
                original time-series)
        rolling: whether to perform time aggregation on a rolling basis or
                whether to use non-overlapping observations

    Returns: numpy arrays of the aggregated log returns from the original
    sample, the aggregated resampled log returns, and the original log returns
    that were not discarded
    """

    # Aggregate log returns in the original sample and obtain a series of
    # retained log returns where the tail end of observations that do not
    # fit an additional block have been cut off.

    aggregated_log_returns, retained_log_returns \
        = aggregate_log_returns(
            log_returns=log_returns,
            number_of_observations_to_aggregate
            =number_of_observations_to_aggregate,
            rolling=rolling,
            verbose=verbose)

    if not sub_sample_length:
        sub_sample_length = len(retained_log_returns)

    # draw sample time-series of log returns of the same length as the
    # original data from retained log returns
    if block_length == 1:
        resampled_log_returns \
            = iid_bootstrap(retained_log_returns,
                            replications=replications,
                            sub_sample_length=sub_sample_length)

    elif block_bootstrap_type == BlockBootstrapType.MOVING_BLOCK:
        resampled_log_returns \
            = moving_block_bootstrap(retained_log_returns,
                                     block_length=block_length,
                                     replications=replications,
                                     sub_sample_length=sub_sample_length)

    elif block_bootstrap_type == BlockBootstrapType.CIRCULAR_BLOCK:
        resampled_log_returns \
            = circular_block_bootstrap(retained_log_returns,
                                       block_length=block_length,
                                       replications=replications,
                                       sub_sample_length=sub_sample_length)

    elif block_bootstrap_type == BlockBootstrapType.STATIONARY:
        resampled_log_returns \
            = stationary_bootstrap(retained_log_returns,
                                   block_length=block_length,
                                   replications=replications,
                                   sub_sample_length=sub_sample_length)
    elif block_bootstrap_type == BlockBootstrapType.TAPERED_BLOCK:
        resampled_log_returns \
            = tapered_block_bootstrap(retained_log_returns,
                                      block_length=block_length,
                                      replications=replications,
                                      sub_sample_length=sub_sample_length)
    else:
        raise ValueError(
            f'block_bootstrap_type {block_bootstrap_type} is not supported')

    # aggregate log-returns in each bootstrapped sample
    aggregated_resampled_log_returns, _ \
        = aggregate_log_returns(
            log_returns=resampled_log_returns,
            number_of_observations_to_aggregate
            =number_of_observations_to_aggregate,
            rolling=rolling,
            verbose=verbose)

    # return the aggregated returns from the bootstrap,
    # the aggregated log returns from the original data,
    # and the retained log returns which the bootstrap is based on
    return (aggregated_resampled_log_returns,
            aggregated_log_returns,
            retained_log_returns)


def resample_and_aggregate_multidimensional_log_returns(
        log_returns: np.ndarray,
        block_length: tp.Union[
            int, float],
        replications: int,
        number_of_observations_to_aggregate: int,
        block_bootstrap_type: BlockBootstrapType
        = BlockBootstrapType.STATIONARY) \
        -> np.ndarray:
    """
    This function resamples a multi-dimensional time-series of log returns and
    sums the resampled returns in blocks.

    Args:
        log_returns: a numpy array of log returns with shape (T, dimension)
        block_length: block length for bootstrap
        replications: the number of bootstrap samples to generate
        number_of_observations_to_aggregate: the number of returns to aggregate
        block_bootstrap_type: The block bootstrap method to use
                              (see class BlockBootstrapType)

    Returns: a numpy array of dimension (replications,
                                         number_of_blocks,
                                         dimension)

    """

    T, dim = log_returns.shape
    number_of_complete_blocks \
        = calculate_number_of_complete_blocks(
            sample_length=T,
            block_length=number_of_observations_to_aggregate)
    number_of_observations_to_retain \
        = number_of_complete_blocks * number_of_observations_to_aggregate

    if block_length == 1:
        resampled_log_returns \
            = iid_bootstrap(log_returns,
                            replications=replications,
                            sub_sample_length=number_of_observations_to_retain)

    elif block_bootstrap_type == BlockBootstrapType.MOVING_BLOCK:
        resampled_log_returns \
            = moving_block_bootstrap(
                log_returns,
                block_length=block_length,
                replications=replications,
                sub_sample_length=number_of_observations_to_retain)

    elif block_bootstrap_type == BlockBootstrapType.CIRCULAR_BLOCK:
        resampled_log_returns \
            = circular_block_bootstrap(
                log_returns,
                block_length=block_length,
                replications=replications,
                sub_sample_length=number_of_observations_to_retain)

    elif block_bootstrap_type == BlockBootstrapType.STATIONARY:
        resampled_log_returns \
            = stationary_bootstrap(
                log_returns,
                block_length=block_length,
                replications=replications,
                sub_sample_length=number_of_observations_to_retain)
    elif block_bootstrap_type == BlockBootstrapType.TAPERED_BLOCK:
        # resampled_log_returns \
        #     = tapered_block_bootstrap(
        #         log_returns,
        #         block_length=block_length,
        #         replications=replications,
        #         sub_sample_length=number_of_observations_to_retain)
        resampled_log_returns \
            = tapered_block_bootstrap_vectorized(
                log_returns,
                block_length=block_length,
                replications=replications,
                sub_sample_length=number_of_observations_to_retain)
    else:
        raise ValueError(
            f'block_bootstrap_type {block_bootstrap_type} is not supported')

    aggregated_resampled_log_returns \
        = aggregate_multi_dimensional_log_returns(
            log_returns=resampled_log_returns,
            number_of_observations_to_aggregate
            =number_of_observations_to_aggregate)

    return aggregated_resampled_log_returns



import numba
import numpy as np
import typing as tp

from recombinator.numba_rng_tools import rng_link


def _construct_kwargs(x: np.ndarray,
                     number_of_boostrap_replications: int,
                     block_length: tp.Optional[int] = None) \
        -> tp.Dict[str, tp.Any]:
    kwargs = {'x': x,
              'replications': number_of_boostrap_replications,
              'link_rngs': True}
    if block_length is not None:
        kwargs['block_length'] = block_length

    return kwargs


@rng_link()
@numba.njit
def numba_rand(link_rngs: bool):
    return np.random.rand()


def numpy_to_numba_rng_link_generic_tester(
        x: np.ndarray,
        bootstrap_function: tp.Callable,
        number_of_boostrap_replications: int,
        block_length: tp.Optional[int] = None):
    kwargs = \
        _construct_kwargs(x=x,
                          number_of_boostrap_replications
                          =number_of_boostrap_replications,
                          block_length=block_length)

    np.random.seed(42)
    resampled_numba_1 = \
        bootstrap_function(**kwargs)

    np.random.seed(42)
    resampled_numba_2 = \
        bootstrap_function(**kwargs)

    np.random.seed(42)
    np.random.rand()
    resampled_numba_3 = \
        bootstrap_function(**kwargs)

    assert np.allclose(resampled_numba_1, resampled_numba_2)
    assert not np.allclose(resampled_numba_1, resampled_numba_3)


def numba_to_numpy_rng_link_generic_tester(
        x: np.ndarray,
        bootstrap_function: tp.Callable,
        number_of_boostrap_replications: int,
        block_length: tp.Optional[int] = None):
    kwargs = \
        _construct_kwargs(x=x,
                          number_of_boostrap_replications
                          =number_of_boostrap_replications,
                          block_length=block_length)

    np.random.seed(42)
    numpy_rand_1 = np.random.rand()

    np.random.seed(42)
    bootstrap_function(**kwargs)

    numpy_rand_2 = np.random.rand()

    assert not np.allclose(numpy_rand_1, numpy_rand_2)



from .iid_bootstrap import \
    iid_bootstrap, \
    iid_bootstrap_via_choice, \
    iid_bootstrap_vectorized, \
    iid_bootstrap_with_antithetic_resampling, \
    iid_balanced_bootstrap, \
    iid_bootstrap_via_loop

from .block_bootstrap import \
    moving_block_bootstrap_vectorized, \
    moving_block_bootstrap, \
    circular_block_bootstrap_vectorized, \
    circular_block_bootstrap, \
    stationary_bootstrap

from .tapered_block_bootstrap import \
    tapered_block_bootstrap_vectorized, \
    tapered_block_bootstrap



import numpy as np
import typing as tp


def estimate_bias_from_bootstrap(
        bootstrap_estimates: np.ndarray,
        original_estimate: float) \
        -> float:
    """
    This function estimates the bias of an estimator given an estimate of a
    statistic from the original data as well as variety of estimates of the same
    statistic from resampled data.

    Args:
        bootstrap_estimates:
            a NumPy array of dimension (number_of_replications, ) containing
            the statistic computed from resampled data
        original_estimate: the statistic computed from the original data
    """

    bootstrap_estimate_of_bias \
        = np.mean(bootstrap_estimates) - original_estimate
    return np.asscalar(bootstrap_estimate_of_bias)


def estimate_standard_error_from_bootstrap(
        bootstrap_estimates: np.ndarray,
        original_estimate: float,
        ddof: int = 0) \
        -> float:
    """
    This function estimates the standard error of an estimator given an estimate
    of a statistic from the original data as well as variety of estimates of the
    same statistic from resampled data.

    Args:
        bootstrap_estimates:
            a NumPy array of dimension (number_of_replications, ) containing
            the statistic computed from resampled data
        original_estimate: the statistic computed from the original data
        ddof: Delta degrees of freedom
    """

    B = len(bootstrap_estimates)
    bootstrap_estimate_of_variance \
        = np.sum((bootstrap_estimates - original_estimate) ** 2) / (B - ddof)
    bootstrap_estimate_of_standard_error \
        = np.sqrt(bootstrap_estimate_of_variance)

    return bootstrap_estimate_of_standard_error


def estimate_confidence_interval_from_bootstrap(
        bootstrap_estimates: np.ndarray,
        confidence_level: float = 95.0) \
        -> tp.Tuple[float, float]:
    """
    This function estimates a confidence interval of an estimator given a
    variety of estimates of the same statistic from resampled data.

    Args:
        bootstrap_estimates: a NumPy array of dimension (B, ) containing the statistic
                    computed from resampled data
        confidence_level: the confidence level associated with the confidence
                          interval in percent (i.e. between 0 and 100)
    """

    percent = 100.0 - confidence_level
    bootstrap_confidence_interval \
        = (np.percentile(bootstrap_estimates, percent / 2.0),
           np.percentile(bootstrap_estimates, 100.0 - percent / 2.0))

    return bootstrap_confidence_interval



import numpy as np
import pytest


@pytest.fixture()
def original_iid_sample_size() -> int:
    return 100


@pytest.fixture()
def original_time_series_sample_size() -> int:
    return 1000


@pytest.fixture()
def number_of_boostrap_replications() -> int:
    return 100000


@pytest.fixture()
def number_of_time_series_boostrap_replications() -> int:
    return 10000


@pytest.fixture()
def iid_sample(original_iid_sample_size: int) -> np.ndarray:
    n = original_iid_sample_size
    np.random.seed(1)
    x = np.abs(np.random.randn(n))

    return x


@pytest.fixture()
def time_series_sample(original_time_series_sample_size: int) -> np.ndarray:
    np.random.seed(1)

    # number of time periods
    T = original_time_series_sample_size

    # draw random errors
    e = np.random.randn(T)
    y = np.zeros((T,))

    # y is an AR(1) with phi_1 = 0.5
    phi_1 = 0.5
    y[0] = e[0] * np.sqrt(1.0 / (1.0 - phi_1 ** 2))
    for t in range(1, T):
        y[t] = phi_1 * y[t - 1] + e[t]

    return y



import math
import numpy as np
import typing as tp

from statsmodels.tsa.ar_model import AR

from recombinator.block_bootstrap import moving_block_bootstrap
from recombinator.block_bootstrap import moving_block_bootstrap_vectorized
from recombinator.block_bootstrap import circular_block_bootstrap
from recombinator.block_bootstrap import circular_block_bootstrap_vectorized
from recombinator.block_bootstrap import stationary_bootstrap
from recombinator.optimal_block_length import optimal_block_length
from recombinator.tapered_block_bootstrap import tapered_block_bootstrap
from recombinator.tapered_block_bootstrap import \
    tapered_block_bootstrap_vectorized

from recombinator.statistics import (
    estimate_confidence_interval_from_bootstrap,
    estimate_standard_error_from_bootstrap
)

from recombinator.tests.rng_link_tools import (
    numpy_to_numba_rng_link_generic_tester,
    numba_to_numpy_rng_link_generic_tester
)


def _test_block_bootstrap_generic(
        bootstrap_function: tp.Callable,
        y: np.ndarray,
        block_length: tp.Union[int, float],
        replications: int,
        lower_ci_bounds: tp.Tuple[float, float] = (0.35, 0.38),
        upper_ci_bounds: tp.Tuple[float, float] = (0.48, 0.50),
        bse_bounds: tp.Tuple[float, float] = (0.05, 0.065),
        test_rng_link: bool = False):
    T = len(y)
    # original estimate
    ar = AR(y)
    estimate_from_original_data = ar.fit(maxlag=1)

    y_star \
        = bootstrap_function(x=y,
                             block_length=block_length,
                             replications=replications)

    assert y_star.shape == (replications, T)

    estimates_from_bootstrap = []
    ar_estimates_from_bootstrap = np.zeros((len(y_star),))

    for b in range(len(y_star)):
        y_bootstrap = np.array(y_star[b, :].squeeze())
        ar_bootstrap = AR(y_bootstrap)
        estimate_from_bootstrap = ar_bootstrap.fit(maxlag=1)
        estimates_from_bootstrap.append(estimate_from_bootstrap)
        ar_estimates_from_bootstrap[b] = estimate_from_bootstrap.params[1]

    lower_ci, upper_ci \
        = estimate_confidence_interval_from_bootstrap(
            bootstrap_estimates=ar_estimates_from_bootstrap,
            confidence_level=95)

    assert lower_ci_bounds[0] <= lower_ci <= lower_ci_bounds[1]
    assert upper_ci_bounds[0] <= upper_ci <= upper_ci_bounds[1]

    bootstrap_standard_error \
        = estimate_standard_error_from_bootstrap(
            bootstrap_estimates=ar_estimates_from_bootstrap,
            original_estimate=estimate_from_original_data.params[1])

    assert bse_bounds[0] <= bootstrap_standard_error <= bse_bounds[1]

    if test_rng_link:
        numpy_to_numba_rng_link_generic_tester(
            x=y,
            bootstrap_function=bootstrap_function,
            number_of_boostrap_replications=replications,
            block_length=block_length)

        numba_to_numpy_rng_link_generic_tester(
            x=y,
            bootstrap_function=bootstrap_function,
            number_of_boostrap_replications=replications,
            block_length=block_length)


def test_moving_block_bootstrap(
        time_series_sample: np.ndarray,
        number_of_time_series_boostrap_replications: int):
    y = time_series_sample
    B = number_of_time_series_boostrap_replications
    b_star = optimal_block_length(y)
    block_length = math.ceil(b_star[0].b_star_cb)

    lower_ci_bounds = (0.35, 0.38)
    upper_ci_bounds = (0.48, 0.50)
    bse_bounds = (0.03, 0.07)

    _test_block_bootstrap_generic(
        bootstrap_function=moving_block_bootstrap,
        y=time_series_sample,
        block_length=block_length,
        replications=number_of_time_series_boostrap_replications,
        lower_ci_bounds=lower_ci_bounds,
        upper_ci_bounds=upper_ci_bounds,
        bse_bounds=bse_bounds,
        test_rng_link=True)

    _test_block_bootstrap_generic(
        bootstrap_function=moving_block_bootstrap_vectorized,
        y=time_series_sample,
        block_length=block_length,
        replications=number_of_time_series_boostrap_replications,
        lower_ci_bounds=lower_ci_bounds,
        upper_ci_bounds=upper_ci_bounds,
        bse_bounds=bse_bounds)


def test_circular_block_bootstrap(
        time_series_sample: np.ndarray,
        number_of_time_series_boostrap_replications: int):
    y = time_series_sample
    B = number_of_time_series_boostrap_replications
    b_star = optimal_block_length(y)
    block_length = math.ceil(b_star[0].b_star_cb)

    lower_ci_bounds = (0.34, 0.37)
    upper_ci_bounds = (0.46, 0.50)
    bse_bounds = (0.03, 0.07)

    _test_block_bootstrap_generic(
        bootstrap_function=circular_block_bootstrap,
        y=time_series_sample,
        block_length=block_length,
        replications=number_of_time_series_boostrap_replications,
        lower_ci_bounds=lower_ci_bounds,
        upper_ci_bounds=upper_ci_bounds,
        bse_bounds=bse_bounds,
        test_rng_link=True)

    _test_block_bootstrap_generic(
        bootstrap_function=circular_block_bootstrap_vectorized,
        y=time_series_sample,
        block_length=block_length,
        replications=number_of_time_series_boostrap_replications,
        lower_ci_bounds=lower_ci_bounds,
        upper_ci_bounds=upper_ci_bounds,
        bse_bounds=bse_bounds)


def test_stationary_bootstrap(
        time_series_sample: np.ndarray,
        number_of_time_series_boostrap_replications: int):
    y = time_series_sample
    B = number_of_time_series_boostrap_replications
    b_star = optimal_block_length(y)
    block_length = b_star[0].b_star_sb

    lower_ci_bounds = (0.35, 0.38)
    upper_ci_bounds = (0.46, 0.50)
    bse_bounds = (0.03, 0.07)

    _test_block_bootstrap_generic(
        bootstrap_function=stationary_bootstrap,
        y=time_series_sample,
        block_length=block_length,
        replications=number_of_time_series_boostrap_replications,
        lower_ci_bounds=lower_ci_bounds,
        upper_ci_bounds=upper_ci_bounds,
        bse_bounds=bse_bounds,
        test_rng_link=True)


def test_stationary_bootstrap_source_array_dimensions(
        time_series_sample,
        original_time_series_sample_size
):
    print('time_series_sample.shape')
    print(time_series_sample.shape)

    bootstrap_result = stationary_bootstrap(
        time_series_sample,
        block_length=800,
        # block_length=10,
        replications=10000,
        sub_sample_length=2 * time_series_sample.shape[0]
    )


def test_tapered_block_bootstrap(
        time_series_sample: np.ndarray,
        number_of_time_series_boostrap_replications: int):
    y = time_series_sample
    B = number_of_time_series_boostrap_replications
    b_star = optimal_block_length(y)
    block_length = math.ceil(b_star[0].b_star_cb)

    lower_ci_bounds = (0.38, 0.42)
    upper_ci_bounds = (0.52, 0.54)
    bse_bounds = (0.03, 0.07)

    _test_block_bootstrap_generic(
        bootstrap_function=tapered_block_bootstrap,
        y=time_series_sample,
        block_length=block_length,
        replications=number_of_time_series_boostrap_replications,
        lower_ci_bounds=lower_ci_bounds,
        upper_ci_bounds=upper_ci_bounds,
        bse_bounds=bse_bounds,
        test_rng_link=True)

    _test_block_bootstrap_generic(
        bootstrap_function=tapered_block_bootstrap_vectorized,
        y=time_series_sample,
        block_length=block_length,
        replications=number_of_time_series_boostrap_replications,
        lower_ci_bounds=lower_ci_bounds,
        upper_ci_bounds=upper_ci_bounds,
        bse_bounds=bse_bounds)



import cocos.numerics as cn
import typing as tp


from .block_bootstrap import (
    moving_block_bootstrap_vectorized,
    circular_block_bootstrap_vectorized
)

from .iid_bootstrap import (
    iid_bootstrap_vectorized,
    iid_bootstrap_via_choice,
    iid_bootstrap_with_antithetic_resampling
)

from .tapered_block_bootstrap import tapered_block_bootstrap_vectorized


def iid_bootstrap_cocos(x: cn.ndarray,
                        replications: int,
                        sub_sample_length: tp.Optional[int] = None) \
        -> cn.ndarray:
    """
    This function performs an i.i.d. non-balanced bootstrap via
    a vectorized implementation.

    Args:
        x: (n,) or (n, k) dimensional Cocos array of input data
        replications: the number of samples to generate
        sub_sample_length: the length of the bootstrapped samples to generate

    Returns:
        a (replications, n) or (replications, n, k) dimensional Cocos array
    """

    return iid_bootstrap_vectorized(x=x,
                                    replications=replications,
                                    sub_sample_length=sub_sample_length)


def iid_bootstrap_via_choice_cocos(
            x: cn.ndarray,
            replications: int,
            sub_sample_length: tp.Optional[int] = None) \
        -> cn.ndarray:
    """
    This function performs an i.i.d. non-balanced bootstrap via
    NumPy's random choice function.

    Args:
        x: (n,) dimensional Cocos array of input data
        replications: the number of samples to generate
        sub_sample_length: the length of the bootstrapped samples to generate

    Returns:
        a (replications, n) dimensional Cocos array
    """

    return iid_bootstrap_via_choice(x=x,
                                    replications=replications,
                                    sub_sample_length=sub_sample_length)


def iid_bootstrap_with_antithetic_resampling_cocos(
            x: cn.ndarray,
            replications: int,
            sub_sample_length: tp.Optional[int] = None) \
        -> cn.ndarray:
    """
    This function performs an i.i.d. non-balanced bootstrap with antithetic
    sampling.

    Args:
        x: (n,) dimensional Cocos array of input data
        replications: the number of samples to generate
        sub_sample_length: the length of the bootstrapped samples to generate

    Returns:
        a (replications, n) dimensional Cocos array
    """

    return iid_bootstrap_with_antithetic_resampling(
                x=x,
                replications=replications,
                sub_sample_length=sub_sample_length,
                num_pack=cn,
                randint=cn.random.randint)


def circular_block_bootstrap_cocos(
        x: cn.ndarray,
        block_length: int,
        replications: int,
        sub_sample_length: tp.Optional[int] = None,
        replace: bool = True) \
        -> cn.ndarray:
    """
    This function creates sub-samples from a data series via circular block
    bootstrapping.

    Args:
        x: Input data, a Cocos array with one or two axes. If y has two axes,
           the Txk array is interpreted as a multidimensional time-series with T
           observations and k variables.
        block_length: the block size
        replications: the number of (sub-)samples to generate
        sub_sample_length: length of the sub-samples to generate
        replace: whether to sample the same block more than once in a single
                 replication

    Returns: a Cocos array with shape (replications, sub_sample_length) of
             bootstrapped sub-samples
    """

    return circular_block_bootstrap_vectorized(
                x=x,
                block_length=block_length,
                replications=replications,
                sub_sample_length=sub_sample_length,
                replace=replace,
                num_pack=cn,
                choice=cn.random.choice)


def moving_block_bootstrap_cocos(
        x: cn.ndarray,
        block_length: int,
        replications: int,
        sub_sample_length: tp.Optional[int] = None,
        replace: bool = True) \
        -> cn.ndarray:
    """
    This function creates sub-samples from a data series via moving block
    bootstrapping. It relies on a vectorized implementation.

    Args:
        x: Input data, a Cocos array with one or two axes. If y has two axes,
           the Txk array is interpreted as a multidimensional time-series with T
           observations and k variables.
        block_length: the block size
        replications: the number of (sub-)samples to generate
        sub_sample_length: length of the sub-samples to generate
        replace: whether to sample the same block more than once in a single
                 replication


    Returns: a Cocos array with shape (replications, sub_sample_length) of
             bootstrapped sub-samples
    """
    return moving_block_bootstrap_vectorized(
                x=x,
                block_length=block_length,
                replications=replications,
                sub_sample_length=sub_sample_length,
                replace=replace,
                num_pack=cn,
                choice=cn.random.choice)


def tapered_block_bootstrap_cocos(
        x: cn.ndarray,
        block_length: int,
        replications: int,
        sub_sample_length: tp.Optional[int] = None,
        replace: bool = True) \
        -> cn.ndarray:
    """
    This function creates samples from a data series using the tapered block
    bootstrap. It samples overlapping blocks of fixed length. Observations are
    first demeaned and then rescaled according to a window function. After the
    simulation, the mean is added back to each observation.

    Args:
        x: Input data, a Cocos array with one or two axes. If y has two axes,
           the Txk array is interpreted as a multidimensional time-series with T
           observations and k variables.
        block_length: the block length
        replications: the number of bootstrap samples to generate
        sub_sample_length: length of the sub-samples to generate
        replace: whether to sample the same block more than once in a single
                 replication

    Returns: a Cocos array with shape (replications, int(len(x)/block_length))
             of bootstrapped sub-samples

    """

    return tapered_block_bootstrap_vectorized(
                    x=x,
                    block_length=block_length,
                    replications=replications,
                    sub_sample_length=sub_sample_length,
                    replace=replace)



import numba
import numpy as np
import typing as tp

from .numba_rng_tools import rng_link
from .utilities import (
    _verify_shape_of_bootstrap_input_data_and_get_dimensions,
    _grab_sub_samples_from_indices,
    _verify_block_bootstrap_arguments,
    BlockBootstrapType,
    _generate_block_start_indices_and_successive_indices
)


@numba.njit
def trapezoid(t: float, c: float) -> float:
    """
    Function computes a trapezoid on [0, 1]
    Args:
        t: real valued argument
        c: parameter indicating the beginning of the flat part

    Returns:
        function value of the trapezoid with parameter c evaluated at t

    """
    if 0 <= t < c:
        return t/c
    elif c <= t < 1-c:
        return 1
    elif 1-c <= t < 1:
        return (1-t)/c
    else:
        return 0.0


@numba.njit
def compute_weights(block_length: int, c: float = 0.43) \
        -> tp.Tuple[np.ndarray, np.ndarray]:
    """
    This function computes the weights to rescale the samples in a tapered block
    bootstrap from the demeaned time-series. The unscaled weights represent the
    naive rescaling. The rescaled weights are chosen such that the variance of
    the bootstrapped  time-series matches that of the data and should be used
    for rescaling the bootstrapped samples.
    The procedure assumes a trapezoidal window.

    Args:
        block_length: the block length
        c: partameter describing the beginning of the flat part of the
           trapezoid.

    Returns:
        a Tuple of unscaled_weights and scaled_weights

    """
    js = list(range(1, block_length+1))
    unscaled_weights \
        = np.array([trapezoid((j - 0.5)/block_length, c) for j in js])
    l2_norm_of_weights = np.sqrt(np.sum(unscaled_weights**2))
    scaled_weights \
        = unscaled_weights * np.sqrt(block_length) / l2_norm_of_weights
    return unscaled_weights, scaled_weights


@rng_link()
@numba.njit
def _tapered_block_bootstrap_internal(block_length: int,
                                      replications: int,
                                      sub_sample_length: int,
                                      T: int,
                                      k: int,
                                      y: np.ndarray,
                                      replace: bool,
                                      link_rngs: bool) \
        -> np.ndarray:
    # compute the number of blocks (k in the original paper)
    number_of_blocks = int(np.ceil(sub_sample_length / block_length))

    # length of the output samples
    output_length = number_of_blocks * block_length

    # compute weights from window function
    unscaled_weights, scaled_weights = compute_weights(block_length)

    # allocate output array
    y_star = np.zeros((replications, output_length, k), dtype=y.dtype)

    block_start_indices = np.arange(np.int32(T - block_length))

    for b in range(replications):
        for m in range(number_of_blocks):
            u = np.random.choice(block_start_indices, replace=replace)
            for j in range(block_length):
                scaled_weight = scaled_weights[j]
                unweighted_observation = y[u + j, :]
                weighted_observation = scaled_weight * unweighted_observation
                y_star[b, m * block_length + j, :] = weighted_observation

    return y_star


def tapered_block_bootstrap(
        x: np.ndarray,
        block_length: int,
        replications: int,
        sub_sample_length: tp.Optional[int] = None,
        replace: bool = True,
        link_rngs: bool = True) \
        -> np.ndarray:
    """
    This function creates samples from a data series using the tapered block
    bootstrap. It samples overlapping blocks of fixed length. Observations are
    first demeaned and then rescaled according to a window function. After the
    simulation, the mean is added back to each observation.

    Args:
        x: input data, a NumPy array with one axis
        block_length: the block length
        replications: the number of bootstrap samples to generate
        sub_sample_length: length of the sub-samples to generate
        replace: whether to sample the same block more than once in a single
                 replication
        link_rngs: whether to synchronize the states of Numba's and Numpy's
                   random number generators

    Returns: a NumPy array with shape (replications, int(len(x)/block_length))
             of bootstrapped sub-samples
    """

    # length of the data sample
    # T = len(x)
    T, k = _verify_shape_of_bootstrap_input_data_and_get_dimensions(x)

    if sub_sample_length is None:
        sub_sample_length = T

    _verify_block_bootstrap_arguments(
        x=x,
        block_length=block_length,
        replications=replications,
        replace=replace,
        bootstrap_type=BlockBootstrapType.TAPERED_BLOCK,
        sub_sample_length=sub_sample_length)

    if block_length > T:
        raise ValueError(
            'The argument block_length must not exceed the size of the data.')

    # compute the mean of x
    x_bar = np.mean(x, axis=0)

    # demean observations
    y = x - x_bar
    if k == 1:
        y = y.reshape((T, k, 1))

    # compute the number of blocks (k in the original paper)
    number_of_blocks = int(np.ceil(sub_sample_length / block_length))

    # length of the output samples
    output_length = number_of_blocks * block_length

    y_star = _tapered_block_bootstrap_internal(
                block_length=block_length,
                replications=replications,
                sub_sample_length=sub_sample_length,
                T=T,
                k=k,
                y=y,
                replace=replace,
                link_rngs=link_rngs)

    if k == 1:
        y_star = y_star.reshape((replications, output_length))

    # add back the mean
    x_star = y_star + x_bar

    # cut off observations exceeding the sub-sample length
    x_star = x_star[:, :sub_sample_length]

    return x_star


def tapered_block_bootstrap_vectorized(
        x: np.ndarray,
        block_length: int,
        replications: int,
        sub_sample_length: tp.Optional[int] = None,
        replace: bool = True,
        num_pack=np,
        choice=np.random.choice) \
        -> np.ndarray:
    """
    This function creates samples from a data series using the tapered block
    bootstrap. It samples overlapping blocks of fixed length. Observations are
    first demeaned and then rescaled according to a window function. After the
    simulation, the mean is added back to each observation.

    Args:
        x: Input data, a NumPy array with one or two axes. If y has two axes,
           the Txk array is interpreted as a multidimensional time-series with T
           observations and k variables.
        block_length: the block length
        replications: the number of bootstrap samples to generate
        sub_sample_length: length of the sub-samples to generate
        replace: whether to sample the same block more than once in a single
                 replication
        num_pack: a module compatible with NumPy (function uses vstack and sort)
        choice: a function compatible with numpy.random.choice

    Returns: a NumPy array with shape (replications, int(len(x)/block_length))
             of bootstrapped sub-samples

    """
    T, k = _verify_shape_of_bootstrap_input_data_and_get_dimensions(x)

    if sub_sample_length is None:
        sub_sample_length = T

    _verify_block_bootstrap_arguments(
                x=x,
                block_length=block_length,
                replications=replications,
                replace=replace,
                bootstrap_type=BlockBootstrapType.TAPERED_BLOCK,
                sub_sample_length=sub_sample_length)

    if block_length > T:
        raise ValueError(
            'The argument block_length must not exceed the size of the data.')

    # compute the mean of x
    x_bar = num_pack.mean(x, axis=0)

    # demean observations
    y = x - x_bar

    # compute weights from window function
    unscaled_weights, scaled_weights = compute_weights(block_length)

    # compute the number of blocks (k in the original paper)
    number_of_blocks = int(np.ceil(sub_sample_length / block_length))

    # replicate scaled weights
    replicated_scaled_weights = num_pack.tile(num_pack.array(scaled_weights),
                                              number_of_blocks)

    if k == 1:
        replicated_scaled_weights = replicated_scaled_weights.reshape((1, -1))
    else:
        replicated_scaled_weights \
            = replicated_scaled_weights.reshape((1, -1, 1))

    block_start_indices = num_pack.array(np.arange(np.int32(T - block_length)))
    # generate a 1-d array containing the sequence of integers from
    # 0 to block_length-1 with shape (1, block_length)
    successive_indices \
        = num_pack.array(np.arange(block_length, dtype=int)
                         .reshape((1, 1, block_length)))

    # generate a random array of block start indices with shape
    # (np.ceil(T/block_length), 1)
    indices = choice(block_start_indices,
                     size=(number_of_blocks,
                           replications,
                           1),
                     replace=replace)

    # add successive indices to starting indices
    indices = (indices + successive_indices)

    # transform to col vector and and remove excess
    indices = indices.reshape((replications, -1))

    # y-star sample simulation
    y_star = _grab_sub_samples_from_indices(y, indices)
    y_star *= replicated_scaled_weights

    # cut off observations exceeding the sub-sample length
    y_star = y_star[:, :sub_sample_length]

    # add back the mean
    x_star = y_star + x_bar

    return x_star



import numpy as np
import typing as tp


from recombinator.iid_bootstrap import \
    iid_balanced_bootstrap, \
    iid_bootstrap, \
    iid_bootstrap_vectorized, \
    iid_bootstrap_via_choice, \
    iid_bootstrap_via_loop, \
    iid_bootstrap_with_antithetic_resampling

from recombinator.statistics import \
    estimate_confidence_interval_from_bootstrap, \
    estimate_standard_error_from_bootstrap

from recombinator.tests.rng_link_tools import (
    numpy_to_numba_rng_link_generic_tester,
    numba_to_numpy_rng_link_generic_tester
)


def _test_iid_bootstrap_generic(bootstrap_function: tp.Callable,
                                x: np.ndarray,
                                replications: int,
                                test_rng_link: bool):
    n = len(x)
    percentile = 75
    original_statistic = np.percentile(x, percentile)

    x_resampled \
        = bootstrap_function(x, replications=replications)

    assert x_resampled.shape == (replications, n)

    resampled_statistic \
        = np.percentile(x_resampled,
                        percentile,
                        axis=1)

    estimated_standard_error \
        = estimate_standard_error_from_bootstrap(
            bootstrap_estimates=resampled_statistic,
            original_estimate=original_statistic)

    assert 0.10 <= estimated_standard_error <= 0.12

    lower_ci, upper_ci \
        = estimate_confidence_interval_from_bootstrap(
            bootstrap_estimates=resampled_statistic,
            confidence_level=95)

    assert 0.82 <= lower_ci <= 0.86
    assert 1.1 <= upper_ci <= 1.2

    if test_rng_link:
        numpy_to_numba_rng_link_generic_tester(
            x=x,
            bootstrap_function=bootstrap_function,
            number_of_boostrap_replications=replications)

        numba_to_numpy_rng_link_generic_tester(
            x=x,
            bootstrap_function=bootstrap_function,
            number_of_boostrap_replications=replications)


def test_iid_bootstrap(iid_sample: np.ndarray,
                       number_of_boostrap_replications):
    _test_iid_bootstrap_generic(
        bootstrap_function=iid_bootstrap,
        x=iid_sample,
        replications=number_of_boostrap_replications,
        test_rng_link=True)


def test_iid_bootstrap_via_loop(iid_sample: np.ndarray,
                                number_of_boostrap_replications):
    _test_iid_bootstrap_generic(
        bootstrap_function=iid_bootstrap_via_loop,
        x=iid_sample,
        replications=number_of_boostrap_replications,
        test_rng_link=True)


def test_iid_bootstrap_vectorized(iid_sample: np.ndarray,
                                  number_of_boostrap_replications):
    _test_iid_bootstrap_generic(
        bootstrap_function=iid_bootstrap_vectorized,
        x=iid_sample,
        replications=number_of_boostrap_replications,
        test_rng_link=False)


def test_iid_bootstrap_via_choice(iid_sample: np.ndarray,
                                  number_of_boostrap_replications):
    _test_iid_bootstrap_generic(
        bootstrap_function=iid_bootstrap_via_choice,
        x=iid_sample,
        replications=number_of_boostrap_replications,
        test_rng_link=False)


def test_iid_balanced_bootstrap(iid_sample: np.ndarray,
                                number_of_boostrap_replications):
    _test_iid_bootstrap_generic(
        bootstrap_function=iid_balanced_bootstrap,
        x=iid_sample,
        replications=number_of_boostrap_replications,
        test_rng_link=False)


def test_iid_antithetic_bootstrap(iid_sample: np.ndarray,
                                  number_of_boostrap_replications):
    _test_iid_bootstrap_generic(
        bootstrap_function=iid_bootstrap_with_antithetic_resampling,
        x=iid_sample,
        replications=number_of_boostrap_replications,
        test_rng_link=False)



import math
import numba
import numpy as np
import typing as tp

from .numba_rng_tools import rng_link
from .utilities import (
    _verify_shape_of_bootstrap_input_data_and_get_dimensions,
    _grab_sub_samples_from_indices
)


def _verify_iid_bootstrap_arguments(
                x: np.ndarray,
                replications: int,
                replace: bool,
                sub_sample_length: tp.Optional[int] = None) -> None:
    T, _ = _verify_shape_of_bootstrap_input_data_and_get_dimensions(x)

    if not sub_sample_length:
        sub_sample_length = T

    if replications <= 0:
        raise ValueError('The argument replications must be strictly positive.')

    if sub_sample_length <= 0:
        raise ValueError('The argument sub_sample_length must be strictly '
                         'positive.')

    if sub_sample_length > T:
        if not replace:
            raise ValueError(
                'The argument sub_sample_length must not exceed the size of '
                'the data.')


# ------------------------------------------------------------------------------
# Non-Balanced Bootstrap via Numba Loops
# - Separate Functions for One-Dimensional and Multi-Dimensional Data
@rng_link()
@numba.njit
def _iid_bootstrap_loop_one_dimensional(
        x: np.ndarray,
        replications: int,
        sub_sample_length: int,
        replace: bool,
        link_rngs: bool) -> np.ndarray:
    n = len(x)

    # allocate memory for resampled data
    x_star = np.empty((replications, sub_sample_length),
                      dtype=x.dtype)

    if replace:
        # randint is faster than choice
        for b in range(replications):
            # generate random integer indices from 0 to n-1
            indices = np.random.randint(low=0,
                                  high=n,
                                  size=(sub_sample_length,))

            # x-star sample simulation
            x_star[b, :] = x[indices]
    else:
        for b in range(replications):
            x_star[b, :] = np.random.choice(x, size=(sub_sample_length,),
                                            replace=replace)

    return x_star


# ToDo: Unify one-dimensional and multi-dimensional bootstrap numba functions
@rng_link()
@numba.njit
def _iid_bootstrap_loop_multi_dimensional(
        x: np.ndarray,
        n: int,
        k: int,
        replications: int,
        sub_sample_length: int,
        replace: bool,
        link_rngs: bool) -> np.ndarray:
    # allocate memory for samples
    x_star = np.empty((replications, sub_sample_length, k),
                      dtype=x.dtype)

    if replace:
        # loop over replications
        for b in range(replications):
            # generate random integer indices from 0 to n-1
            indices = np.random.randint(low=0,
                                        high=n,
                                        size=(sub_sample_length,))

            x_star[b, :, :] = x[indices, :]
    else:
        # loop over replications
        for b in range(replications):
            # generate random integer indices from 0 to n-1
            indices = np.random.choice(np.arange(n),
                                       size=(sub_sample_length,),
                                       replace=replace)

            x_star[b, :, :] = x[indices, :]

    return x_star


def _iid_bootstrap_via_loop_one_dimensional(
        x: np.ndarray,
        replications: int,
        sub_sample_length: tp.Optional[int] = None,
        replace: bool = True,
        link_rngs: bool = True) -> np.ndarray:
    """
    This function performs an i.i.d. non-balanced bootstrap using loops in Numba
    for one-dimensional input data.

    Args:
        x: (n,) dimensional NumPy array of input data
        replications: the number of samples to generate
        sub_sample_length: the length of the bootstrapped samples to generate
        replace: whether to repeat the same observation in a given sub-sample
        link_rngs: whether to synchronize the states of Numba's and Numpy's
                   random number generators

    Returns:
        a (replications, n) dimensional NumPy array
    """

    # dimensions of original data
    n, k = _verify_shape_of_bootstrap_input_data_and_get_dimensions(x)

    if k > 1:
        raise ValueError(
            'This function only supports one-dimensional input data.')

    if not sub_sample_length:
        sub_sample_length = n

    _verify_iid_bootstrap_arguments(
        x,
        replications=replications,
        replace=replace,
        sub_sample_length=sub_sample_length)

    x_star = _iid_bootstrap_loop_one_dimensional(
                            x=x,
                            replications=replications,
                            sub_sample_length=sub_sample_length,
                            replace=replace,
                            link_rngs=link_rngs)

    assert x_star.shape == (replications, n)

    return x_star


def _iid_bootstrap_via_loop_multi_dimensional(
        x: np.ndarray,
        replications: int,
        sub_sample_length: tp.Optional[int] = None,
        replace: bool = True,
        link_rngs: bool = True) -> np.ndarray:
    """
    This function performs an i.i.d. non-balanced bootstrap using loops in Numba
    for multi-dimensional input data.

    Args:
        x: (n, k) dimensional NumPy array of input data
        replications: the number of samples to generate
        sub_sample_length: the length of the bootstrapped samples to generate
        replace: whether to repeat the same observation in a given sub-sample
        link_rngs: whether to synchronize the states of Numba's and Numpy's
                   random number generators

    Returns:
        a (replications, n, k) dimensional NumPy array
    """

    # dimensions of original data
    n, k = _verify_shape_of_bootstrap_input_data_and_get_dimensions(x)

    if k == 1:
        raise ValueError('This function does not support one-dimensional data.')

    if not sub_sample_length:
        sub_sample_length = n

    _verify_iid_bootstrap_arguments(
        x,
        replications=replications,
        replace=replace,
        sub_sample_length=sub_sample_length)

    # simulate samples
    x_star = _iid_bootstrap_loop_multi_dimensional(
                            x=x,
                            n=n,
                            k=k,
                            replications=replications,
                            sub_sample_length=sub_sample_length,
                            replace=replace,
                            link_rngs=link_rngs)

    assert x_star.shape == (replications, sub_sample_length, k)

    return x_star


def iid_bootstrap_via_loop(x: np.ndarray,
                           replications: int,
                           sub_sample_length: tp.Optional[int] = None,
                           link_rngs: bool = True) \
        -> np.ndarray:
    """
    This function performs an i.i.d. non-balanced bootstrap using loops in
    Numba.

    Args:
        x: (n,) or (n, k) dimensional NumPy array of input data
        replications: the number of samples to generate
        sub_sample_length: the length of the bootstrapped samples to generate
        link_rngs: whether to synchronize the states of Numba's and Numpy's
                   random number generators

    Returns:
        a (replications, n) or (replications, n, k) dimensional NumPy array
    """

    # dimensions of original data
    n, k = _verify_shape_of_bootstrap_input_data_and_get_dimensions(x)

    if not sub_sample_length:
        sub_sample_length = n

    _verify_iid_bootstrap_arguments(
        x,
        replications=replications,
        replace=True,
        sub_sample_length=sub_sample_length)

    if k == 1:
        return _iid_bootstrap_via_loop_one_dimensional(
                                x=x,
                                replications=replications,
                                sub_sample_length=sub_sample_length,
                                link_rngs=link_rngs)

    else:
        return _iid_bootstrap_via_loop_multi_dimensional(
                                x=x,
                                replications=replications,
                                sub_sample_length=sub_sample_length,
                                link_rngs=link_rngs)


# ------------------------------------------------------------------------------
# non-balanced bootstrap via choice
def iid_bootstrap_via_choice(x: np.ndarray,
                             replications: int,
                             sub_sample_length: tp.Optional[int] = None,
                             choice=np.random.choice) \
        -> np.ndarray:
    """
    This function performs an i.i.d. non-balanced bootstrap via
    NumPy's random choice function.

    Args:
        x: (n,) dimensional NumPy array of input data
        replications: the number of samples to generate
        sub_sample_length: the length of the bootstrapped samples to generate
        choice: a function compatible with numpy.random.choice

    Returns:
        a (replications, n) dimensional NumPy array
    """

    # dimensions of original data
    n, k = _verify_shape_of_bootstrap_input_data_and_get_dimensions(x)

    if k > 1:
        raise ValueError('iid bootstrap via choice does not support '
                         'multi-dimensional input data')

    if not sub_sample_length:
        sub_sample_length = n

    _verify_iid_bootstrap_arguments(
        x,
        replications=replications,
        replace=True,
        sub_sample_length=sub_sample_length)

    # generate random integer indices from 0 to n-1
    x_star = choice(x, size=(replications, sub_sample_length), replace=True)

    assert x_star.shape == (replications, sub_sample_length)

    return x_star


# ------------------------------------------------------------------------------
# non-balanced bootstrap vectorized
def iid_bootstrap_vectorized(x: np.ndarray,
                             replications: int,
                             sub_sample_length: tp.Optional[int] = None,
                             randint=np.random.randint) \
        -> np.ndarray:
    """
    This function performs an i.i.d. non-balanced bootstrap via
    a vectorized implementation.

    Args:
        x: (n,) or (n, k) dimensional NumPy array of input data
        replications: the number of samples to generate
        sub_sample_length: the length of the bootstrapped samples to generate
        randint: a numpy.random.randint compatible function to generate random
                 integers

    Returns:
        a (replications, n) or (replications, n, k) dimensional NumPy array
    """

    # dimensions of original data
    n, k = _verify_shape_of_bootstrap_input_data_and_get_dimensions(x)

    if not sub_sample_length:
        sub_sample_length = n

    _verify_iid_bootstrap_arguments(
        x,
        replications=replications,
        replace=True,
        sub_sample_length=sub_sample_length)

    # generate random integer indices from 0 to n-1
    indices = randint(low=0, high=n, size=(replications, sub_sample_length))

    # grab samples
    x_star = _grab_sub_samples_from_indices(x, indices)

    if k == 1:
        assert x_star.shape == (replications, sub_sample_length)
    else:
        assert x_star.shape == (replications, sub_sample_length, k)

    return x_star


# ------------------------------------------------------------------------------
# Non-Balanced Bootstrap with Antithetic Resampling - Vectorized Implementation
def iid_bootstrap_with_antithetic_resampling(
                        x: np.ndarray,
                        replications: int,
                        sub_sample_length: tp.Optional[int] = None,
                        num_pack=np,
                        randint=np.random.randint) \
        -> np.ndarray:
    """
    This function performs an i.i.d. non-balanced bootstrap with antithetic
    sampling.

    Args:
        x: (n,) dimensional NumPy array of input data
        replications: the number of samples to generate
        sub_sample_length: the length of the bootstrapped samples to generate
        num_pack: a module compatible with NumPy (function uses vstack and sort)
        randint: a numpy.random.randint compatible function to generate random
                 integers

    Returns:
        a (replications, n) dimensional NumPy array
    """

    # dimensions of original data
    n, k = _verify_shape_of_bootstrap_input_data_and_get_dimensions(x)

    if k > 1:
        raise ValueError('Antithetic resampling is not supported for '
                         'multi-dimensional data.')

    if not sub_sample_length:
        sub_sample_length = n

    _verify_iid_bootstrap_arguments(x,
                                    replications=replications,
                                    replace=True,
                                    sub_sample_length=sub_sample_length)

    # sort original sample
    x = num_pack.sort(x)

    # generate random integer indices from 0 to n-1
    indices_1 = randint(low=0,
                        high=n,
                        size=(math.ceil(replications / 2),
                              sub_sample_length))
    indices_2 = n - indices_1 - 1
    indices = num_pack.vstack((indices_1,
                               indices_2[:math.floor(replications / 2), :]))

    # grab samples
    x_star = _grab_sub_samples_from_indices(x, indices)

    assert x_star.shape == (replications, sub_sample_length)

    return x_star


# ------------------------------------------------------------------------------
# Non-Balanced Bootstrap with Antithetic Resampling - Numba Loop Implementation
@rng_link()
@numba.njit
def _iid_bootstrap(x: np.ndarray,
                   replications: int,
                   sub_sample_length: int,
                   replace: bool,
                   link_rngs: bool) -> np.ndarray:
    n = len(x)

    # allocate memory for indices
    indices = np.empty((replications, sub_sample_length),
                       dtype=np.int32)

    if replace:
        for b in range(replications):
            # generate random integer indices from 0 to n-1
            indices[b, :] = np.random.randint(low=0,
                                              high=n,
                                              size=(sub_sample_length,))

    else:
        for b in range(replications):
            indices[b, :] = np.random.choice(np.arange(n),
                                             size=(sub_sample_length,),
                                             replace=replace)

    return indices


def iid_bootstrap(x: np.ndarray,
                  replications: int,
                  sub_sample_length: tp.Optional[int] = None,
                  replace: bool = True,
                  antithetic: bool = False,
                  link_rngs: bool = True) -> np.ndarray:
    """
    This function performs an i.i.d. non-balanced bootstrap using loops in Numba
    for one-dimensional input data.

    Args:
        x: (n,) dimensional NumPy array of input data
        replications: the number of samples to generate
        sub_sample_length: the length of the bootstrapped samples to generate
        replace: whether to repeat the same observation in a given sub-sample
        antithetic: whether to use antithetic sampling
        link_rngs: whether to synchronize the states of Numba's and Numpy's
                   random number generators

    Returns:
        a (replications, n) dimensional NumPy array
    """

    # dimensions of original data
    n, k = _verify_shape_of_bootstrap_input_data_and_get_dimensions(x)

    if not sub_sample_length:
        sub_sample_length = n

    if k > 1 and antithetic:
        raise ValueError('Antithetic resampling is not supported for '
                         'multi-dimensional data.')

    _verify_iid_bootstrap_arguments(
        x,
        replications=replications,
        replace=replace,
        sub_sample_length=sub_sample_length)

    if antithetic:
        x = np.sort(x)

        indices_1 = _iid_bootstrap(x=x,
                                   replications=math.ceil(replications / 2),
                                   sub_sample_length=sub_sample_length,
                                   replace=replace,
                                   link_rngs=link_rngs)

        indices_2 = n - indices_1 - 1
        indices = np.vstack((indices_1,
                             indices_2[:math.floor(replications / 2), :]))
    else:
        indices = _iid_bootstrap(x=x,
                                 replications=replications,
                                 sub_sample_length=sub_sample_length,
                                 replace=replace,
                                 link_rngs=link_rngs)

    x_star = _grab_sub_samples_from_indices(x, indices)

    if k == 1:
        assert x_star.shape == (replications, sub_sample_length)
    else:
        assert x_star.shape == (replications, sub_sample_length, k)

    return x_star


# ------------------------------------------------------------------------------
# Balanced Bootstrap
def iid_balanced_bootstrap(x: np.ndarray,
                           replications: int,
                           num_pack=np,
                           permutation=np.random.permutation,
                           shuffle=np.random.shuffle) -> np.ndarray:
    """
    This function performs an i.i.d. balanced bootstrap. That means
    each value of the original sample appears 'replication' times
    in all of the bootstrapped samples altogether.

    Args:
        x: (n, k) dimensional NumPy array of input data
        replications: the number of samples to generate
        num_pack: a module compatible with NumPy (function uses vstack and sort)
        permutation: a function compatible with numpy.random.permutation
        shuffle: a function compatible with numpy.random.shuffle

    Returns:
        a (replications, n) or (replications, n, k) dimensional NumPy array
    """

    # dimensions of original data
    n, k = _verify_shape_of_bootstrap_input_data_and_get_dimensions(x)

    _verify_iid_bootstrap_arguments(x,
                                    replications=replications,
                                    replace=True,
                                    sub_sample_length=n)

    # simulate sub-samples
    if k > 1:
        indices = (permutation(num_pack
                               .arange(n)
                               .repeat(replications))
                   .reshape(replications, n))
        x_star = _grab_sub_samples_from_indices(x, indices)
        assert x_star.shape == (replications, n, k)
    else:
        y = num_pack.repeat(x, replications)
        shuffle(y)
        x_star = y.reshape(replications, n)

        assert x_star.shape == (replications, n)

    return x_star



# This module is a Python adaptation of Andrew Patten's Matlab implementation
# available at http://public.econ.duke.edu/~ap172/opt_block_length_REV_dec07.txt


# from dataclasses import dataclass
import numpy as np
import typing as tp


def mlag(x: np.ndarray,
         n: tp.Optional[int] = 1,
         init: tp.Optional[float] = 0.0) -> np.ndarray:
    """
    Purpose: generates a matrix of n lags from a matrix (or vector)
    containing a set of vectors (For use in var routines)

    Usage:     xlag = mlag(x,nlag)
    or: xlag1 = mlag(x), which defaults to 1-lag
    where: x is a nobs by nvar NumPy array

    Args:
        nlag = # of contiguous lags for each vector in x
        init = (optional) scalar value to feed initial missing values
                (default = 0)

    Returns: xlag = a matrix of lags (nobs x nvar*nlag)
    x1(t-1), x1(t-2), ... x1(t-nlag), x2(t-1), ... x2(t-nlag) ...


    original Matlab version written by:
    James P. LeSage, Dept of Economics
    University of Toledo
    2801 W. Bancroft St,
    Toledo, OH 43606
    jpl@jpl.econ.utoledo.edu

    Adapted for Python August 12, 2018 by Michael C. Nowotny
    """

    nobs, nvar = x.shape

    xlag = np.ones((nobs, nvar * n), dtype=x.dtype) * init
    icnt = 0
    for i in range(nvar):
        for j in range(n):
            xlag[j + 1:, icnt + j] = x[0:-j - 1, i]
        icnt += n

    return xlag


def lam(kk: np.ndarray) -> np.ndarray:
    """
    Helper function, calculates the flattop kernel weights.

    Adapted for Python August 12, 2018 by Michael C. Nowotny
    """
    return (np.abs(kk) >= 0) * (np.abs(kk) < 0.5) \
           + 2 * (1.0 - np.abs(kk)) * (np.abs(kk) >= 0.5) * (np.abs(kk) <= 1)


# @dataclass(frozen=True)
class OptimalBlockLength(tp.NamedTuple):
    b_star_sb: float  # optimal block length for stationary bootstrap
    b_star_cb: float  # optimal block length for circular block bootstrap


# ToDo: Add calculation of optimal block length for moving block bootstrap
# ToDo: Add calculation of optimal block length for tapered block bootstrap
def optimal_block_length(data: np.ndarray) -> tp.Sequence[OptimalBlockLength]:
    """
    This is a function to select the optimal (in the sense of minimising the MSE
    of the estimator of the long-run variance) block length for the stationary
    bootstrap or circular bootstrap.
    The code follows Politis and White, 2001,
    "Automatic Block-Length Selection for the Dependent Bootstrap".

        DECEMBER 2007: CORRECTED TO DEAL WITH ERROR IN LAHIRI'S PAPER, PUBLISHED
        BY NORDMAN IN THE ANNALS OF STATISTICS

        NOTE: The optimal average block length for the stationary bootstrap,
              and it does not need to be an integer.
              The optimal block length for the circular bootstrap should be an
              integer. Politis and White suggest rounding the output UP to the
              nearest integer.

     Args:
         data, an nxk matrix

     Returns: a 2xk NumPy array of optimal bootstrap block lengths,
              [[b_star_sb], [b_star_cb]], where
              b_star_sb: optimal block length for stationary bootstrap
              b_star_cb: optimal block length for circular bootstrap

    original Matlab version written by:
    Andrew Patton

    4 December, 2002
    Revised (to include CB): 13 January, 2003.

    Helpful suggestions for this code were received from
    Dimitris Politis and Kevin Sheppard.

    Modified 23.8.2003 by Kevin Sheppard for speed issues

    Adapted for Python August 12, 2018 by Michael C. Nowotny
    """

    if data.ndim == 1:
        data = data.reshape((-1, 1))
    elif data.ndim > 2:
        raise ValueError(
            'data must be a two dimensional NumPy array'
            '(number of observations x number of variables)')
    n, k = data.shape

    # these are optional in the original Matlab implementation
    # opt_block_length_full.m, but fixed at default values here
    kn = int(max(5, np.sqrt(np.log10(n))))

    # adding kn extra lags to employ Politis' (2002) suggestion
    # for finding largest significant m
    m_max = int(np.ceil(np.sqrt(n)) + kn)

    # maximum value of b_star_sb to consider.
    # dec07: new idea for rule-of-thumb to put an upper bound on estimated
    # optimal block length
    b_max = np.ceil(min(3 * np.sqrt(n), n / 3))

    c = 2
    original_data = data
    # b_star_final = np.zeros((2, k), dtype=np.float64)
    b_star_final = []

    for i in range(k):
        data = original_data[:, i].reshape((-1, 1))

        # FIRST STEP: finding m_hat-> the largest lag for which the
        # auto-correlation is still significant.
        temp = mlag(data, m_max)

        # dropping the first m_max rows, as they are filled with zeros
        temp = temp[m_max:, :]
        temp = np.corrcoef(np.hstack((data[m_max:], temp)), rowvar=False)
        temp = temp[1:, 0].reshape((-1, 1))

        # We follow the empirical rule suggested in
        # Politis, 2002, "Adaptive Bandwidth Choice".
        # as suggested in Remark 2.3, setting c=2, kn=5

        # looking at vectors of auto-correlations,
        # from lag m_hat to lag m_hat+kn
        temp2 = np.hstack((np.transpose(mlag(temp, kn)), temp[-kn:]))

        # dropping the first kn-1, as the vectors have empty cells
        temp2 = temp2[:, kn:]

        # checking which are less than the critical value
        temp2 = np.abs(temp2) < (c * np.sqrt(np.log10(n) / n)
                                 * np.ones((kn, m_max - kn + 1)))

        # this counts the number of insignificant autocorrelations
        temp2 = np.sum(temp2, axis=0).reshape((1, -1))
        temp3 = np.hstack((np.arange(1, temp2.shape[1] + 1).reshape((-1, 1)),
                           temp2.transpose()))

        # selecting all rows where ALL kn auto-correlations are not significant
        temp3 = temp3[np.squeeze(temp2 == kn), :]

        if temp3.size == 0:
            # this means that NO collection of kn auto-correlations were all
            # insignificant, so pick largest significant lag
            m_hat = max(
                np.flatnonzero(np.abs(temp) > (c * np.sqrt(np.log10(n) / n))))
        else:
            # if more than one collection is possible, choose the smallest m
            m_hat = temp3[0, 0]

        if 2 * m_hat > m_max:
            m = m_max
        else:
            m = 2 * m_hat

        del temp, temp2, temp3

        # SECOND STEP: computing the inputs to the function for b_star_sb
        kk = np.arange(-m, m + 1).reshape((-1, 1))
        if m > 0:
            temp = mlag(data, m)

            # dropping the first m_max rows, as they're filled with zeros
            temp = temp[m:, :]
            temp = np.cov(np.hstack((data[m:], temp)).transpose())

            # auto-covariances
            acv = temp[:, 0].reshape((-1, 1))
            acv2 = np.hstack(
                (-np.arange(1, m + 1).reshape((-1, 1)), acv[1:, :]))
            if acv2.shape[0] > 1:
                acv2 = acv2[acv2[:, 0].argsort(),]

            # auto-covariances from -m to m
            acv = np.vstack((acv2[:, 1].reshape((-1, 1)), acv))
            del acv2

            g_hat = np.sum(lam(kk / m) * np.abs(kk) * acv)
            dcb_hat = (4.0 / 3.0) * np.sum(lam(kk / m) * acv) ** 2

            # first part of dsb_hat (note cos(0)=1)
            dsb_hat = 2 * (np.sum(lam(kk / m) * acv) ** 2)

            # FINAL STEP: constructing the optimal block length estimator

            # optimal block lenght for stationary bootstrap
            b_star_sb = ((2 * (g_hat ** 2) / dsb_hat) ** (1.0 / 3.0)) \
                     * (n ** (1.0 / 3.0))
            if b_star_sb > b_max:
                b_star_sb = b_max

            # optimal block length for circular bootstrap
            b_star_cb = ((2 * (g_hat ** 2) / dcb_hat) ** (1.0 / 3.0)) \
                        * (n ** (1.0 / 3.0))
            if b_star_cb > b_max:
                b_star_cb = b_max

            # b_star = (b_star_sb, b_star_cb)
            b_star = OptimalBlockLength(b_star_sb=b_star_sb,
                                        b_star_cb=b_star_cb)
        else:
            b_star = OptimalBlockLength(b_star_sb=1.0, b_star_cb=1.0)

        b_star_final.append(b_star)

    return tuple(b_star_final)



import math
import numba
import numpy as np
import typing as tp

from .numba_rng_tools import rng_link
from .utilities import (
    _generate_block_start_indices_and_successive_indices,
    _verify_shape_of_bootstrap_input_data_and_get_dimensions,
    _grab_sub_samples_from_indices,
    _verify_block_bootstrap_arguments,
    BlockBootstrapType
)


# ------------------------------------------------------------------------------
# stationary bootstrap
@rng_link()
@numba.njit
def _stationary_bootstrap_loop(block_length: float,
                               replications: int,
                               sub_sample_length: int,
                               indices: np.ndarray,
                               T: int,
                               link_rngs: bool) -> np.ndarray:
    """
    This function implements the inner loop for stationary bootstraps. It is JIT
    compiled via Numba for performance.

    Args:
        block_length: the average length of each block in a stationary bootstrap
        replications: the number of (sub-)samples to generate
        sub_sample_length: length of the sub-samples to generate
        indices: an integer Numpy array with shape (replications, sub_sample_length)
           containing (random) indices of the the beginning of the first block
           in the first column.
        T: the length of the data series that is sub-sampled
        link_rngs: whether to synchronize the states of Numba's and Numpy's
                   random number generators

    Returns: an integer Numpy array with shape (replications, sub_sample_length)
             containing the index into the source data series for each element
             of each sub-sample.
    """
    limit = 2*T

    for b in range(replications):
        for t in range(1, sub_sample_length):
            if np.random.rand() < 1.0 / block_length:
                # end current block and randomly pick start index for next block
                indices[b, t] = np.ceil(T * np.random.rand())
            else:
                # continue current block for another time-step
                next_index = indices[b, t - 1] + 1

                if next_index >= limit:
                    next_index = 0

                indices[b, t] = next_index

    return indices


def stationary_bootstrap(x: np.ndarray,
                         block_length: float,
                         replications: int,
                         sub_sample_length: tp.Optional[int] = None,
                         link_rngs: bool = True) \
        -> np.ndarray:
    """
    This function creates sub-samples from a data series via stationary
    bootstrapping.

    Args:
        x: Input data, a NumPy array with one or two axes. If y has two axes,
           the Txk array is interpreted as a multidimensional time-series with T
           observations and k variables.
        block_length: the block length
        replications: the number of (sub-)samples to generate
        sub_sample_length: length of the sub-samples to generate
        link_rngs: whether to synchronize the states of Numba's and Numpy's
                   random number generators

    Returns: a NumPy array with shape (replications, sub_sample_length) of
             bootstrapped sub-samples
    """

    _verify_block_bootstrap_arguments(
        x=x,
        block_length=block_length,
        replications=replications,
        replace=True,
        bootstrap_type=BlockBootstrapType.STATIONARY,
        sub_sample_length=sub_sample_length)

    T, _ = _verify_shape_of_bootstrap_input_data_and_get_dimensions(x)

    if not sub_sample_length:
        sub_sample_length = T

    # repeat the input series to allow blocks to wrap around in a circular
    # fashion
    if x.ndim == 1:
        x = np.hstack((x, x))
    else:
        x = np.vstack((x, x))

    # allocate array for the indices into the source array
    indices = np.zeros((replications, sub_sample_length), dtype=np.int)

    # randomly initialize the beginning of the first block in each sub-sample
    indices[:, 0] = np.random.randint(T, size=(replications,))

    # Loop over B bootstraps
    indices = _stationary_bootstrap_loop(block_length=block_length,
                                         replications=replications,
                                         sub_sample_length=sub_sample_length,
                                         indices=indices,
                                         T=T,
                                         link_rngs=link_rngs)

    # return sub-samples
    return _grab_sub_samples_from_indices(x, indices)


# ------------------------------------------------------------------------------
# moving block and circular bootstrap - Numba versions
@rng_link()
@numba.njit
def _general_block_bootstrap_loop(block_length: int,
                                  replications: int,
                                  block_start_indices: np.ndarray,
                                  successive_indices: np.ndarray,
                                  sub_sample_length: tp.Optional[int],
                                  replace: bool,
                                  link_rngs: bool) -> np.ndarray:
    """
    This function implements the inner loop for moving block or circular block
    bootstrapping. It is JIT compiled via Numba for performance.

    Args:
        block_length: the block length
        replications: the number of (sub-)samples to generate
        block_start_indices: a one-dimensional Numpy array containing the first
                             index of each block in into the array y
        successive_indices: an integer Numpy array of shape (1, block_length)
                           containing the range(0, block_length)
        sub_sample_length: length of the sub-samples to generate
        replace: whether to sample the same block more than once in a single
                 replication
        link_rngs: whether to synchronize the states of Numba's and Numpy's
                   random number generators

    Returns: an integer Numpy array with shape (replications, sub_sample_length)
             containing the index into the source data series for each element
             of each sub-sample.

    """
    # generate array of indeces into the original time series that describe
    # the composition of the generate sub-samples
    indices = np.zeros((replications, sub_sample_length), dtype=np.int32)

    # loop over replications bootstraps
    for b in range(replications):
        # generate a random array of block start indices with shape
        # (np.ceil(sub_sample_length/block_length), 1)
        tmp_indices \
            = np.random.choice(block_start_indices,
                               size=(math.ceil(sub_sample_length
                                               / block_length),
                                     1),
                               replace=replace)

        # ToDo: Move that outside the loop
        # add successive indices to starting indices
        tmp_indices = (tmp_indices + successive_indices)

        # transform to col vector and and remove excess
        indices[b, :] = tmp_indices.reshape((-1,))[:sub_sample_length]

    return indices


def _general_block_bootstrap(x: np.ndarray,
                             block_length: int,
                             replications: int,
                             sub_sample_length: tp.Optional[int] = None,
                             replace: bool = True,
                             circular: bool = False,
                             link_rngs: bool = True) -> np.ndarray:
    """
    This function creates sub-samples from a data series via block based
    bootstrapping using either the circular or moving block scheme.

    Args:
        x: Input data, a NumPy array with one or two axes. If y has two axes,
           the Txk array is interpreted as a multidimensional time-series with T
           observations and k variables.
        block_length: the block size
        replications: the number of (sub-)samples to generate
        sub_sample_length: length of the sub-samples to generate
        replace: whether to sample the same block more than once in a single
                 replication
        circular: whether to use circular or moving block bootstrapping
        link_rngs: whether to synchronize the states of Numba's and Numpy's
                   random number generators

    Returns: a NumPy array with shape (replications, sub_sample_length) of
             bootstrapped sub-samples
    """
    T, _ = _verify_shape_of_bootstrap_input_data_and_get_dimensions(x)

    if not sub_sample_length:
        sub_sample_length = T

    if circular:
        bootstrap_type = BlockBootstrapType.CIRCULAR_BLOCK
    else:
        bootstrap_type = BlockBootstrapType.MOVING_BLOCK

    _verify_block_bootstrap_arguments(x=x,
                                      block_length=block_length,
                                      replications=replications,
                                      replace=replace,
                                      bootstrap_type=bootstrap_type,
                                      sub_sample_length=sub_sample_length)

    block_start_indices, successive_indices \
        = _generate_block_start_indices_and_successive_indices(
                    sample_length=T,
                    block_length=block_length,
                    circular=circular,
                    successive_3d=False)

    if circular:
        # replicate time-series for wrap-around
        if x.ndim == 1:
            x = np.hstack((x, x))
        else:
            x = np.vstack((x, x))

    indices \
        = _general_block_bootstrap_loop(block_length=block_length,
                                        replications=replications,
                                        block_start_indices=block_start_indices,
                                        successive_indices=successive_indices,
                                        sub_sample_length=sub_sample_length,
                                        replace=replace,
                                        link_rngs=link_rngs)

    # return sub-samples
    return _grab_sub_samples_from_indices(x, indices)


def moving_block_bootstrap(x: np.ndarray,
                           block_length: int,
                           replications: int,
                           sub_sample_length: tp.Optional[int] = None,
                           replace: bool = True,
                           link_rngs: bool = True) -> np.ndarray:
    """
    This function creates sub-samples from a data series via moving block
    bootstrapping.

    Args:
        x: Input data, a NumPy array with one or two axes. If y has two axes,
           the Txk array is interpreted as a multidimensional time-series with T
           observations and k variables.
        block_length: the block size
        replications: the number of (sub-)samples to generate
        sub_sample_length: length of the sub-samples to generate
        replace: whether to sample the same block more than once in a single
                 replication
        link_rngs: whether to synchronize the states of Numba's and Numpy's
                   random number generators

    Returns: a NumPy array with shape (replications, sub_sample_length) of
             bootstrapped sub-samples
    """

    return _general_block_bootstrap(x=x,
                                    block_length=block_length,
                                    replications=replications,
                                    sub_sample_length=sub_sample_length,
                                    replace=replace,
                                    circular=False,
                                    link_rngs=link_rngs)


def circular_block_bootstrap(x: np.ndarray,
                             block_length: int,
                             replications: int,
                             sub_sample_length: tp.Optional[int] = None,
                             replace: bool = True,
                             link_rngs: bool = True) -> np.ndarray:
    """
    This function creates sub-samples from a data series via circular block
    bootstrapping.

    Args:
        x: Input data, a NumPy array with one or two axes. If y has two axes,
           the Txk array is interpreted as a multidimensional time-series with T
           observations and k variables.
        block_length: the block size
        replications: the number of (sub-)samples to generate
        sub_sample_length: length of the sub-samples to generate
        replace: whether to sample the same block more than once in a single
                 replication
        link_rngs: whether to synchronize the states of Numba's and Numpy's
                   random number generators

    Returns: a NumPy array with shape (replications, sub_sample_length) of
             bootstrapped sub-samples
    """

    return _general_block_bootstrap(x=x,
                                    block_length=block_length,
                                    replications=replications,
                                    sub_sample_length=sub_sample_length,
                                    replace=replace,
                                    circular=True,
                                    link_rngs=link_rngs)


def _generalized_block_bootstrap_vectorized(
        x: np.ndarray,
        block_length: int,
        replications: int,
        circular: bool,
        sub_sample_length: tp.Optional[int] = None,
        replace: bool = True,
        num_pack=np,
        choice=np.random.choice) \
        -> np.ndarray:
    """
    This function creates sub-samples from a data series via circular block
    bootstrapping.

    Args:
        x: Input data, a NumPy array with one or two axes. If y has two axes,
           the Txk array is interpreted as a multidimensional time-series with T
           observations and k variables.
        block_length: the block size
        replications: the number of (sub-)samples to generate
        sub_sample_length: length of the sub-samples to generate
        replace: whether to sample the same block more than once in a single
                 replication
        num_pack: a module compatible with NumPy (function uses vstack and sort)
        choice: a function compatible with numpy.random.choice

    Returns: a NumPy array with shape (replications, sub_sample_length) of
             bootstrapped sub-samples
    """

    T, _ = _verify_shape_of_bootstrap_input_data_and_get_dimensions(x)

    if not sub_sample_length:
        sub_sample_length = T

    if circular:
        bootstrap_type = BlockBootstrapType.CIRCULAR_BLOCK
    else:
        bootstrap_type = BlockBootstrapType.MOVING_BLOCK

    _verify_block_bootstrap_arguments(
                x=x,
                block_length=block_length,
                replications=replications,
                replace=replace,
                bootstrap_type=bootstrap_type,
                sub_sample_length=sub_sample_length)

    if circular:
        # replicate time-series for wrap-around
        if x.ndim == 1:
            x = num_pack.hstack((x, x))
        else:
            x = num_pack.vstack((x, x))

    block_start_indices, successive_indices \
        = _generate_block_start_indices_and_successive_indices(
                    sample_length=T,
                    block_length=block_length,
                    circular=circular,
                    successive_3d=True)

    # print(f'block_start_indices.shape before = {block_start_indices.shape}')
    # print(f'successive_indices.shape before = {successive_indices.shape}')
    block_start_indices = num_pack.array(block_start_indices)
    successive_indices = num_pack.array((successive_indices))
    # print(f'block_start_indices.shape after = {block_start_indices.shape}')
    # print(f'successive_indices.shape after = {successive_indices.shape}')

    # generate a random array of block start indices with shape
    # (np.ceil(T/block_length), 1)
    indices = choice(block_start_indices,
                     size=(math.ceil(sub_sample_length
                                     / block_length),
                           replications,
                           1),
                     replace=replace)

    # add successive indices to starting indices
    indices = (indices + successive_indices)
    # print(indices)
    # print(f'indices.shape = {indices.shape}')

    # transform to col vector and and remove excess
    indices = indices.reshape((replications, -1))[:, :sub_sample_length]
    # print(indices)

    # return sub-samples
    return _grab_sub_samples_from_indices(x, indices)


def circular_block_bootstrap_vectorized(
        x: np.ndarray,
        block_length: int,
        replications: int,
        sub_sample_length: tp.Optional[int] = None,
        replace: bool = True,
        num_pack=np,
        choice=np.random.choice) \
        -> np.ndarray:
    """
    This function creates sub-samples from a data series via circular block
    bootstrapping.

    Args:
        x: Input data, a NumPy array with one or two axes. If y has two axes,
           the Txk array is interpreted as a multidimensional time-series with T
           observations and k variables.
        block_length: the block size
        replications: the number of (sub-)samples to generate
        sub_sample_length: length of the sub-samples to generate
        replace: whether to sample the same block more than once in a single
                 replication
        num_pack: a module compatible with NumPy (function uses vstack and sort)
        choice: a function compatible with numpy.random.choice

    Returns: a NumPy array with shape (replications, sub_sample_length) of
             bootstrapped sub-samples
    """

    return _generalized_block_bootstrap_vectorized(
                x=x,
                block_length=block_length,
                replications=replications,
                circular=True,
                sub_sample_length=sub_sample_length,
                replace=replace,
                num_pack=num_pack,
                choice=choice)


def moving_block_bootstrap_vectorized(
        x: np.ndarray,
        block_length: int,
        replications: int,
        sub_sample_length: tp.Optional[int] = None,
        replace: bool = True,
        num_pack=np,
        choice=np.random.choice) \
        -> np.ndarray:
    """
    This function creates sub-samples from a data series via moving block
    bootstrapping. It relies on a vectorized implementation.

    Args:
        x: Input data, a NumPy array with one or two axes. If y has two axes,
           the Txk array is interpreted as a multidimensional time-series with T
           observations and k variables.
        block_length: the block size
        replications: the number of (sub-)samples to generate
        sub_sample_length: length of the sub-samples to generate
        replace: whether to sample the same block more than once in a single
                 replication
        num_pack: a module compatible with NumPy (function uses vstack and sort)
        choice: a function compatible with numpy.random.choice


    Returns: a NumPy array with shape (replications, sub_sample_length) of
             bootstrapped sub-samples
    """

    return _generalized_block_bootstrap_vectorized(
                x=x,
                block_length=block_length,
                replications=replications,
                circular=False,
                sub_sample_length=sub_sample_length,
                replace=replace,
                num_pack=num_pack,
                choice=choice)



from enum import Enum
from fractions import Fraction
import math
import numpy as np
import typing as tp


def calculate_number_of_blocks(sample_length: int,
                               block_length: int,
                               overhang: bool = False) -> int:
    """
    This function calculates the number of blocks contained
    in a sample with a given block length.

    Args:
        sample_length: the lenght of the sample to be split into blocks
        block_length: the length of each block
        overhang: whether the end of the sample that would not fit another
                  block should be cut off or combined with elements from
                  the beginning of the sample to form another block.

    Returns: number of blocks
    """

    if overhang:
        return math.ceil(sample_length / block_length)
    else:
        return math.floor(sample_length / block_length)


def number_of_permutations(sub_sample_length: int,
                           block_length: int,
                           number_of_blocks: int,
                           replacement: bool = False) -> int:
    """
    Computes the number of possible permutations of a certain number of
    blocks with a given length in creating sub-samples of a particular
    length if the block order matters.

    Args:
        sub_sample_length: the length of the sub-sample to be generated
        block_length: the length of a single block
        number_of_blocks: the number of blocks to sample from
        replacement: whether to sample a given block more than once

    Returns: number of permutations
    """

    blocks_to_sample = math.ceil(sub_sample_length / block_length)

    if replacement:
        return number_of_blocks ** blocks_to_sample
    else:
        if blocks_to_sample > number_of_blocks:
            raise ValueError(
                f'The number of blocks to sample ({blocks_to_sample}) to '
                f'create a subsample of length {sub_sample_length} without '
                f'replacement exceeds the number of blocks available '
                f'({number_of_blocks}).')

        n_permutations = Fraction(math.factorial(number_of_blocks),
                            math.factorial(number_of_blocks - blocks_to_sample))
        return int(n_permutations)


def number_of_combinations(sub_sample_length: int,
                           block_length: int,
                           number_of_blocks: int,
                           replacement: bool = False) -> int:
    """
    Computes the number of possible combinations of a certain number of
    blocks with a given length in creating sub-samples of a particular
    length if the block order does not matter.

    Args:
        sub_sample_length: the length of the sub-sample to be generated
        block_length: the length of a single block
        number_of_blocks: the number of blocks to sample from
        replacement: whether to sample a given block more than once

    Returns: number of combinations
    """

    blocks_to_sample = math.ceil(sub_sample_length / block_length)

    if replacement:
        n_combinations \
            = Fraction(
                int(math.factorial(blocks_to_sample + number_of_blocks - 1)),
                math.factorial(blocks_to_sample)
                * math.factorial(number_of_blocks - 1))
        return int(n_combinations)
    else:
        n_combinations \
            = Fraction(math.factorial(number_of_blocks),
                       math.factorial(number_of_blocks - blocks_to_sample)
                       * math.factorial(blocks_to_sample))
        return int(n_combinations)


# @numba.njit
def _verify_shape_of_bootstrap_input_data_and_get_dimensions(x: np.ndarray) \
        -> tp.Tuple[int, int]:
    if x.ndim == 1:
        # the input data is a one-dimensional time-series
        T = len(x)
        k = 1
    elif x.ndim == 2:
        # the data is a multi-dimensional time-series
        T, k = x.shape
    else:
        raise ValueError('The argument y must be a NumPy array with one '
                         '(for a univariate time-series) or two axes '
                         '(for a multi-variate time-series).')

    return T, k


def _grab_sub_samples_from_indices(x: np.ndarray,
                                   u: np.ndarray) -> np.ndarray:
    """
    This function selects sub-samples of the data array y at the indices in u.
    There are B replications.
    Args:
        x: a NumPy array of shape (T, k) or (T, )
        u: a NumPy array of shape (B, T) of indices into the data array y

    Returns:

    """

    if x.ndim == 1:
        return x[u]
    else:
        return x[u, :]


class BlockBootstrapType(Enum):
    MOVING_BLOCK = 0
    CIRCULAR_BLOCK = 1
    STATIONARY = 2
    TAPERED_BLOCK = 3


def _verify_block_bootstrap_arguments_internal(
        x: np.ndarray,
        block_length: tp.Union[int, float],
        replications: int,
        replace: bool,
        stationary_bootstrap: bool = False,
        sub_sample_length: tp.Optional[int] = None) -> None:
    """
    This function checks that the arguments passed to block bootstrap functions
    satisfy certain constraints. If a violation is detected, an exception is
    raised.

    Args:
        x: the source data array
        block_length: the length of each block to sample
        replications: the number of samples to bootstrap
        replace: whether blocks can appear more than once in a sample
        stationary_bootstrap: whether this is a stationary bootstrap
        sub_sample_length: the length of each bootstrap sub-sample

    Returns: None

    """
    T, k = _verify_shape_of_bootstrap_input_data_and_get_dimensions(x)

    if stationary_bootstrap:
        if not isinstance(block_length, float) and \
                not isinstance(block_length, int):
            raise ValueError('The block_length must be int or float.')
    else:
        if not isinstance(block_length, int):
            raise ValueError('The block_length must be int.')

    if not sub_sample_length:
        sub_sample_length = T

    if block_length <= 0:
        raise ValueError('The argument block_length must be strictly positive.')

    if replications <= 0:
        raise ValueError('The argument replications must be strictly positive.')

    if sub_sample_length <= 0:
        raise ValueError('The argument sub_sample_length must be strictly '
                         'positive.')

    if block_length > T:
        raise ValueError(
            'The argument block_length must not exceed the size of the data.')

    if sub_sample_length > T:
        if not replace:
        # if not replace or stationary_bootstrap:
            raise ValueError(f'The argument '
                             f'sub_sample_length={sub_sample_length} must not '
                             f'exceed the length of the data={T}.')


def _verify_block_bootstrap_arguments(
        x: np.ndarray,
        block_length: tp.Union[int, float],
        replications: int,
        replace: bool,
        bootstrap_type: BlockBootstrapType,
        sub_sample_length: tp.Optional[int] = None) -> None:
    """
    This function checks that the arguments passed to block bootstrap functions
    satisfy certain constraints. If a violation is detected, an exception is
    raised.

    Args:
        x: the source data array
        block_length: the length of each block to sample
        replications: the number of samples to bootstrap
        replace: whether blocks can appear more than once in a sample
        bootstrap_type: see class BlockBootstrapType
        sub_sample_length: the length of each bootstrap sub-sample

    Returns: None

    """
    stationary_bootstrap = (bootstrap_type is BlockBootstrapType.STATIONARY)
    _verify_block_bootstrap_arguments_internal(
        x,
        block_length=block_length,
        replications=replications,
        replace=replace,
        stationary_bootstrap=stationary_bootstrap,
        sub_sample_length=sub_sample_length)


def _verify_iid_bootstrap_arguments(
                x: np.ndarray,
                replications: int,
                replace: bool,
                sub_sample_length: tp.Optional[int] = None) -> None:
    _verify_block_bootstrap_arguments(
                        x=x,
                        block_length=1,
                        replications=replications,
                        replace=replace,
                        bootstrap_type=BlockBootstrapType.MOVING_BLOCK,
                        sub_sample_length=sub_sample_length)


def _generate_block_start_indices_and_successive_indices(sample_length: int,
                                                         block_length: int,
                                                         circular: bool,
                                                         successive_3d: bool) \
        -> tp.Tuple[np.ndarray, np.ndarray]:
    block_start_indices = list(range(0, sample_length, block_length))
    if not circular:
        if max(block_start_indices) + block_length >= sample_length:
            block_start_indices = block_start_indices[:-1]

    # generate a 1-d array containing the sequence of integers from
    # 0 to block_length-1 with shape (1, block_length)
    if successive_3d:
        successive_indices \
            = np.arange(block_length, dtype=int).reshape((1, 1, block_length))
    else:
        successive_indices \
            = np.arange(block_length, dtype=int).reshape((1, block_length))

    return np.array(block_start_indices), successive_indices


import numpy as np

from recombinator.log_returns import aggregate_multi_dimensional_log_returns

example_array = np.array([[[1, 2], [3, 4], [5, 6], [7, 8]],
                          [[100, 200], [300, 400], [500, 600], [700, 800]]])

expected_aggregated_array \
    = np.array([[[4, 6], [12, 14]], [[400, 600], [1200, 1400]]])


def test_aggregate_multi_dimensional_log_returns():
    aggregated_example \
        = aggregate_multi_dimensional_log_returns(
            log_returns=example_array,
            number_of_observations_to_aggregate=2)

    assert np.allclose(aggregated_example, expected_aggregated_array)



"""
This module is adapted from Alex Gramfort's work <alexandre.gramfort@inria.fr>
published on GitHub at https://github.com/numba/numba/issues/3249.
"""

import numpy as np
from numba import _helperlib


def check_random_state(seed):
    """
    Turn seed into a np.random.RandomState instance.

    If seed is None, return the RandomState singleton used by np.random.
    If seed is an int, return a new RandomState instance seeded with seed.
    If seed is already a RandomState instance, return it.
    Otherwise raise ValueError.
    """
    if seed is None or seed is np.random:
        return np.random.mtrand._rand
    if isinstance(seed, (int, np.integer)):
        return np.random.RandomState(seed)
    if isinstance(seed, np.random.RandomState):
        return seed
    raise ValueError(f'{seed} cannot be used to seed a '
                     f'numpy.random.RandomState instance')


def _copy_np_state(r, ptr):
    """
    Copy state of Numpy random *r* to Numba state *ptr*.
    """
    ints, index = r.get_state()[1:3]
    _helperlib.rnd_set_state(ptr, (index, [int(x) for x in ints]))
    return ints, index


def _copyback_np_state(r, ptr):
    """
    Copy state of of Numba state *ptr* to Numpy random *r*
    """
    index, ints = _helperlib.rnd_get_state(ptr)
    r.set_state(('MT19937', ints, index, 0, 0.0))


def get_np_state_ptr():
    """
    Get the Numba state *ptr*
    """
    return _helperlib.rnd_get_np_state_ptr()


class rng_link:
    """
    This decorator links the state of the random number generators in Numpy and
    Numba. To achieve this, it copies Numpy's random state to Numba before the
    function call and copies the state in the opposite direction afterwards.
    """
    def __init__(self, random_state=None):
        self.random_state = random_state

    def __call__(self, func):
        def new_func(*args, **kwargs):
            link_numba_rng = kwargs.get('link_rngs')
            if link_numba_rng:
                r = check_random_state(self.random_state)
                ptr = get_np_state_ptr()
                _copy_np_state(r, ptr)

            out = func(*args, **kwargs)

            if link_numba_rng:
                _copyback_np_state(r, ptr)

            return out
        return new_func
